from typing import List

from .Expr import *
from .ParserError import ParseError
from .Stmt import *
from .Tokens import TokenType, Token


class Parser:
    def __init__(self, tokens: List[Token]):
        self.tokens = tokens
        self.current = 0

    # expression -> equality ;
    def expression(self) -> Expr:
        return self.assignment()

    # equality -> comparison ( ( "!=" | "==" ) comparison )* ;
    def equality(self) -> Expr:
        expr = self.comparison()

        # 循环相当于 * 的作用，对于 match()  , 相当于 |
        while self.match(TokenType.BANG_EQUAL, TokenType.EQUAL_EQUAL):
            operator = self.previous()
            right = self.comparison()
            expr = Binary(expr, operator, right)

        return expr

    # comparison -> addition ( ( ">" | ">=" | "<" | "<=" ) addition )* ;
    def comparison(self) -> Expr:
        expr = self.addition()

        while self.match(TokenType.GREATER, TokenType.GREATER_EQUAL, TokenType.LESS, TokenType.LESS_EQUAL):
            operator = self.previous()
            right = self.addition()
            expr = Binary(expr, operator, right)

        return expr

    # addition -> multiplication (("-"|"+") multiplication)*;
    def addition(self) -> Expr:
        expr = self.multiplication()

        while self.match(TokenType.MINUS, TokenType.PLUS):
            operator = self.previous()
            right = self.multiplication()
            expr = Binary(expr, operator, right)

        return expr

    # multiplication -> unary (("/"|"*") unary)*;
    def multiplication(self) -> Expr:
        expr = self.unary()

        while self.match(TokenType.SLASH, TokenType.STAR):
            operator = self.previous()
            right = self.unary()
            expr = Binary(expr, operator, right)

        return expr

    # unary -> ("!" | "-"); unary | primary;
    def unary(self) -> Expr:
        if self.match(TokenType.BANG, TokenType.MINUS):
            operator = self.previous()
            right = self.unary()
            return Unary(operator, right)
        else:
            return self.call()

    def call(self):
        expr = self.primary()
        while True:
            if self.match(TokenType.LEFT_PAREN):
                expr = self.finishCall(expr)  # 对可调用对象求值
            else:
                break
        return expr

    def primary(self) -> Expr:
        if self.match(TokenType.FALSE): return Literal(False)
        if self.match(TokenType.TRUE): return Literal(True)
        if self.match(TokenType.NIL): return Literal(None)

        if self.match(TokenType.NUMBER, TokenType.STRING):
            return Literal(self.previous().literal)

        if self.match(TokenType.IDENTIFIER):
            return Variable(self.previous())

        if self.match(TokenType.LEFT_PAREN):
            expr = self.expression()
            self.consume(TokenType.RIGHT_PAREN, "Expect ')' after expression")
            return Grouping(expr)

        raise self.error(self.peek(), "Except expression.")

    # infrastructure primitive operation
    def match(self, *types: TokenType) -> bool:
        """checking to see if the current token is any of the given types"""

        # 只要有一个匹配就返回真
        for type in types:
            if self.check(type):
                self.advance()
                return True
        return False

    def consume(self, type: TokenType, message: str) -> Token:
        """match, advance and raise parser error"""
        if self.check(type):
            return self.advance()
        else:
            raise self.error(self.peek(), message)

    def error(self, token: Token, message: str):
        err = ParseError(token, message)
        err.report()
        return err

    def check(self, tokenType: TokenType) -> bool:
        if self.isAtEnd(): return False  # REVIEW: 注意此处是否定
        return self.peek().type == tokenType  # REVIEW: 再次注意 match 和 peek 的区别，match 通过 advance 消耗 token

    def advance(self) -> Token:
        if not self.isAtEnd(): self.current += 1
        return self.previous()

    def isAtEnd(self) -> bool:
        return self.peek().type == TokenType.EOF

    def peek(self) -> Token:
        return self.tokens[self.current]

    def previous(self) -> Token:
        return self.tokens[self.current - 1]

    def sychronize(self) -> None:
        self.advance()

        while not self.isAtEnd():  # REVIEW: 不止一次忘记的 not
            if self.previous().type == TokenType.SEMICOLON: return

            if self.peek().type in [TokenType.CLASS,
                                    TokenType.FUN,
                                    TokenType.VAR,
                                    TokenType.FUN,
                                    TokenType.IF,
                                    TokenType.WHILE,
                                    TokenType.PRINT,
                                    TokenType.RETURN]:
                return
        self.advance()

    def parse(self) -> List[Stmt]:
        statements = []
        while not self.isAtEnd():
            statements.append(self.declaration())

        return statements

    def statement(self) -> Stmt:
        if (self.match(TokenType.FOR)): return self.forStatement()
        if (self.match(TokenType.IF)): return self.ifStatement()
        if (self.match(TokenType.PRINT)): return self.printStatement()
        if (self.match(TokenType.RETURN)): return self.returnStatement()
        if (self.match(TokenType.WHILE)): return self.whileStatement()
        if (self.match(TokenType.LEFT_BRACE)): return Block(self.block())
        return self.expressionStatement()

    def printStatement(self) -> Stmt:
        value = self.expression()
        self.consume(TokenType.SEMICOLON, "Except ';' after value.")
        return Print(value)

    def expressionStatement(self) -> Stmt:
        """expression statement is a expression with a semicolon end"""
        expr = self.expression()
        self.consume(TokenType.SEMICOLON, "Except ';' after expression.")
        return Expression(expr)

    def block(self) -> List[Stmt]:
        statements = []
        while not self.check(TokenType.RIGHT_BRACE) and not self.isAtEnd():
            statements.append(self.declaration())

        self.consume(TokenType.RIGHT_BRACE, "Except '}' after block.")
        return statements

    def assignment(self):
        expr = self.or_()

        if self.match(TokenType.EQUAL):
            equals = self.previous()
            value = self.assignment()

            if isinstance(expr, Variable):
                name = expr.name
                return Assign(name, value)
            self.error(equals, "Invalid assignment target.")

        return expr

    def declaration(self):
        try:
            if self.match(TokenType.FUN): return self.function('function')
            if self.match(TokenType.VAR): return self.varDeclaration()
            return self.statement()
        except ParseError as error:
            self.sychronize()
            return None

    def varDeclaration(self):
        name = self.consume(TokenType.IDENTIFIER, "Except variable name.")

        initializer = None
        if self.match(TokenType.EQUAL):
            initializer = self.expression()

        self.consume(TokenType.SEMICOLON, "Expcet ';' after variable declaration")
        return Var(name, initializer)

    def ifStatement(self):
        self.consume(TokenType.LEFT_PAREN, "Except '(' after if.")
        condition = self.expression()
        self.consume(TokenType.RIGHT_PAREN, "Except ')' after if conditon.")

        theBranch = self.statement()
        elseBranch = None
        if self.match(TokenType.ELSE):
            elseBranch = self.statement()

        return If(condition, theBranch, elseBranch)

    def or_(self):
        expr = self.and_()

        while self.match(TokenType.OR):
            operator = self.previous()
            right = self.and_()
            expr = Logical(expr, operator, right)

        return expr

    def and_(self):
        expr = self.equality()

        while self.match(TokenType.AND):
            operator = self.previous()
            right = self.equality()
            expr = Logical(expr, operator, right)

        return expr

    def whileStatement(self):
        self.consume(TokenType.LEFT_PAREN, "Except '(' after 'while'.")
        condition = self.expression()
        self.consume(TokenType.RIGHT_PAREN, "Except ') after condition.")
        body = self.statement()

        return While(condition, body)

    def forStatement(self):
        self.consume(TokenType.LEFT_PAREN, "Except '(' after 'for'.")

        if self.match(TokenType.SEMICOLON):
            initializer = None
        elif self.match(TokenType.VAR):
            initializer = self.varDeclaration()
        else:
            initializer = self.expressionStatement()

        condition = None
        if not self.check(TokenType.SEMICOLON):
            condition = self.expression()
        self.consume(TokenType.SEMICOLON, "Except ';' after loop condition.")

        increment = None
        if not self.check(TokenType.RIGHT_PAREN):
            increment = self.expression()
        self.consume(TokenType.RIGHT_PAREN, "Excpet ';' after for caluses.")

        body = self.statement()

        if increment is not None:
            body = Block([body, Expression(increment)])

        if condition is None:
            condition = Literal(True)

        body = While(condition, body)

        if initializer is not None:
            body = Block([initializer, body])

        return body

    def finishCall(self, callee: Expr):
        arguments = []
        if not self.check(TokenType.RIGHT_PAREN):
            arguments.append(self.expression())
            while self.match(TokenType.COMMA):
                if len(arguments) >= 8: self.error(self.peek(), "Cannot have more than 8 arguments.")
                arguments.append(self.expression())
        paren = self.consume(TokenType.RIGHT_PAREN, "Expect ')' after arguments.")
        return Call(callee, paren, arguments)

    def function(self, kind: str) -> Function:
        name = self.consume(TokenType.IDENTIFIER, f"Expect '(' after {kind} name.")
        self.consume(TokenType.LEFT_PAREN, f"Expect '(' after {kind} name.")
        parameters: List[Token] = []
        if not self.check(TokenType.RIGHT_PAREN):
            while True:
                if len(parameters) >= 8:
                    self.error(self.peek(), "Cannot have more than 8 parameters.")
                parameters.append(self.consume(TokenType.IDENTIFIER, "Except parameter name."))
                if not self.match(TokenType.COMMA):
                    break  # REVIEW: 用 while if break 模拟 do while 时注意 if condition 是否定的
        self.consume(TokenType.RIGHT_PAREN, "Except ')' after parameters.")

        self.consume(TokenType.LEFT_BRACE, "Expect '{{' before {} body.".format(kind))
        body = self.block()
        return Function(name, parameters, body)

    def returnStatement(self):
        keyword = self.previous()
        value = None
        if not self.check(TokenType.SEMICOLON):
            value = self.expression()

        self.consume(TokenType.SEMICOLON, "Expect ';' after return value.")
        return Return(keyword, value)



from .Tokens import Token
from typing import List

__all__ = [
    "ExprVisitor",
    "Expr",
    "Assign",
    "Binary",
    "Call",
    "Grouping",
    "Literal",
    "Logical",
    "Unary",
    "Variable",
]

class ExprVisitor:
    def visitAssignExpr(self, expr: "Assign"): raise NotImplementedError

    def visitBinaryExpr(self, expr: "Binary"): raise NotImplementedError

    def visitCallExpr(self, expr: "Call"): raise NotImplementedError

    def visitGroupingExpr(self, expr: "Grouping"): raise NotImplementedError

    def visitLiteralExpr(self, expr: "Literal"): raise NotImplementedError

    def visitLogicalExpr(self, expr: "Logical"): raise NotImplementedError

    def visitUnaryExpr(self, expr: "Unary"): raise NotImplementedError

    def visitVariableExpr(self, expr: "Variable"): raise NotImplementedError


class Expr:
    def accept(self, visitor: ExprVisitor): raise NotImplementedError



class Assign(Expr):

    def __init__(self, name: Token, value: Expr):
        self.name = name
        self.value = value

    def accept(self, visitor: ExprVisitor):
        return visitor.visitAssignExpr(self)


class Binary(Expr):

    def __init__(self, left: Expr, operator: Token, right: Expr):
        self.left = left
        self.operator = operator
        self.right = right

    def accept(self, visitor: ExprVisitor):
        return visitor.visitBinaryExpr(self)


class Call(Expr):

    def __init__(self, callee: Expr, paren: Token, arguments: List[Expr]):
        self.callee = callee
        self.paren = paren
        self.arguments = arguments

    def accept(self, visitor: ExprVisitor):
        return visitor.visitCallExpr(self)


class Grouping(Expr):

    def __init__(self, expression: Expr):
        self.expression = expression

    def accept(self, visitor: ExprVisitor):
        return visitor.visitGroupingExpr(self)


class Literal(Expr):

    def __init__(self, value: object):
        self.value = value

    def accept(self, visitor: ExprVisitor):
        return visitor.visitLiteralExpr(self)


class Logical(Expr):

    def __init__(self, left: Expr, operator: Token, right: Expr):
        self.left = left
        self.operator = operator
        self.right = right

    def accept(self, visitor: ExprVisitor):
        return visitor.visitLogicalExpr(self)


class Unary(Expr):

    def __init__(self, operator: Token, right: Expr):
        self.operator = operator
        self.right = right

    def accept(self, visitor: ExprVisitor):
        return visitor.visitUnaryExpr(self)


class Variable(Expr):

    def __init__(self, name: Token):
        self.name = name

    def accept(self, visitor: ExprVisitor):
        return visitor.visitVariableExpr(self)


    


from . import ErrorState


class ScanError(Exception):
    def __init__(self, line: int, message: str):
        self.line = line
        self.message = message

    def report(self):
        print(f'[line {self.line}] Error: {self.message}')
        ErrorState.hadError = True



# usr/bin env python3
# coding: utf-8

from .Tokens import Token, TokenType
from . import ErrorState


class ParseError(Exception):
    def __init__(self, token: Token, message: str):
        self.token = token
        self.message = message

    def report(self):
        if self.token.type == TokenType.EOF:
            print(f'[line {self.token.line}] Error at end: {self.message}')
        else:
            where = self.token.lexeme
            print(f"[line {self.token.line}] Error at '{where}': {self.message}")

        ErrorState.hadError = True



from .Tokens import Token
from typing import List

__all__ = [
    "StmtVisitor",
    "Stmt",
    "Block",
    "If",
    "Function",
    "Expression",
    "Print",
    "Return",
    "Var",
    "While",
]

class StmtVisitor:
    def visitBlockStmt(self, stmt: "Block"): raise NotImplementedError

    def visitIfStmt(self, stmt: "If"): raise NotImplementedError

    def visitFunctionStmt(self, stmt: "Function"): raise NotImplementedError

    def visitExpressionStmt(self, stmt: "Expression"): raise NotImplementedError

    def visitPrintStmt(self, stmt: "Print"): raise NotImplementedError

    def visitReturnStmt(self, stmt: "Return"): raise NotImplementedError

    def visitVarStmt(self, stmt: "Var"): raise NotImplementedError

    def visitWhileStmt(self, stmt: "While"): raise NotImplementedError


class Stmt:
    def accept(self, visitor: StmtVisitor): raise NotImplementedError



class Block(Stmt):

    def __init__(self, statements: List[Stmt]):
        self.statements = statements

    def accept(self, visitor: StmtVisitor):
        return visitor.visitBlockStmt(self)


class If(Stmt):

    def __init__(self, condition: 'Expr', thenBranch: Stmt, elseBranch: Stmt):
        self.condition = condition
        self.thenBranch = thenBranch
        self.elseBranch = elseBranch

    def accept(self, visitor: StmtVisitor):
        return visitor.visitIfStmt(self)


class Function(Stmt):

    def __init__(self, name: Token, parameters: List[Token], body: List[Stmt]):
        self.name = name
        self.parameters = parameters
        self.body = body

    def accept(self, visitor: StmtVisitor):
        return visitor.visitFunctionStmt(self)


class Expression(Stmt):

    def __init__(self, expression: 'Expr'):
        self.expression = expression

    def accept(self, visitor: StmtVisitor):
        return visitor.visitExpressionStmt(self)


class Print(Stmt):

    def __init__(self, expression: 'Expr'):
        self.expression = expression

    def accept(self, visitor: StmtVisitor):
        return visitor.visitPrintStmt(self)


class Return(Stmt):

    def __init__(self, keyword: Token, value: 'Expr'):
        self.keyword = keyword
        self.value = value

    def accept(self, visitor: StmtVisitor):
        return visitor.visitReturnStmt(self)


class Var(Stmt):

    def __init__(self, name: Token, initializer: 'Expr'):
        self.name = name
        self.initializer = initializer

    def accept(self, visitor: StmtVisitor):
        return visitor.visitVarStmt(self)


class While(Stmt):

    def __init__(self, condition: 'Expr', body: Stmt):
        self.condition = condition
        self.body = body

    def accept(self, visitor: StmtVisitor):
        return visitor.visitWhileStmt(self)


    


from typing import List

from . import Stmt
from .Environment import Environment
from .LoxCallable import LoxCallable
from .Return import Return


class LoxFunction(LoxCallable):
    def __init__(self, declaration: Stmt.Function, closure: Environment):
        self.declaration = declaration
        self.closure = closure

    def call(self, interpreter: "Interpreter", arguments: List[object]):
        environment = Environment(self.closure)
        for i in range(len(self.declaration.parameters)):
            environment.define(self.declaration.parameters[i].lexeme, arguments[i])

        try:
            interpreter.executeBlock(self.declaration.body, environment)
        except Return as returnValue:
            return returnValue.value

    def arity(self):
        return len(self.declaration.parameters)

    def __str__(self):
        return f"<function {self.declaration.name.lexeme} >"



from typing import List

import time

from . import Expr
from . import Stmt
from .Environment import Environment
from .LoxCallable import LoxCallable
from .LoxFunction import LoxFunction
from .Return import Return
from .Tokens import TokenType, Token
from .LoxRuntimeError import LoxRuntimeError, runtimeError


class Interpreter(Expr.ExprVisitor, Stmt.StmtVisitor):
    def __init__(self):
        self.globals = Environment()
        self.environment = self.globals
        # 用 type 仿照 Java 建立匿名类
        self.globals.define("clock", type('AnonymousClass', (LoxCallable,),
                                          {'arity': lambda self: 0,
                                           'call': lambda self, interpreter, arguments: time.perf_counter()})())

    def visitLiteralExpr(self, expr: Expr.Literal) -> object:
        return expr.value

    def visitLogicalExpr(self, expr: Expr.Logical):
        """short-circuits and return"""
        left = self.evaluate(expr.left)
        if expr.operator.type == TokenType.OR:  # or
            if self.isTruthy(left): return left
        else:  # and
            if not self.isTruthy(left): return left

        return self.evaluate(expr.right)

    def visitUnaryExpr(self, expr: Expr.Unary):
        right = self.evaluate(expr.right)

        if expr.operator.type == TokenType.MINUS:
            self.checkNumberOperand(expr.operator, right)
            return -float(right)
        elif expr.operator == TokenType.BANG:
            return not self.isTruthy(right)

        # Unreachable
        return None

    def visitVariableExpr(self, expr: Expr.Variable):
        return self.environment.get(expr.name)

    # REVIEW: 这里用默认参数来替代实现 Java 的 Overloading，看起来不太漂亮
    def checkNumberOperand(self, operator: Token, left: object, right: object = None):
        if right is None:
            if isinstance(left, float): return
            raise LoxRuntimeError(operator, "Operand must be number")
        else:
            if isinstance(left, float) and isinstance(right, float): return
            raise LoxRuntimeError(operator, "Operator must be number.")

    def visitGroupingExpr(self, expr: Expr.Grouping) -> object:
        return self.evaluate(expr.expression)

    def visitBinaryExpr(self, expr: Expr.Binary):
        left = self.evaluate(expr.left)
        right = self.evaluate(expr.right)

        optype = expr.operator.type
        if optype == TokenType.BANG_EQUAL:
            return not self.isEqual(left, right)
        elif optype == TokenType.EQUAL_EQUAL:
            return self.isEqual(left, right)
        elif optype == TokenType.GREATER:
            self.checkNumberOperand(expr.operator, left, right)
            return float(left) > float(right)
        elif optype == TokenType.GREATER_EQUAL:
            self.checkNumberOperand(expr.operator, left, right)
            return float(left) >= float(right)
        elif optype == TokenType.LESS:
            self.checkNumberOperand(expr.operator, left, right)
            return float(left) < float(right)
        elif optype == TokenType.LESS_EQUAL:
            self.checkNumberOperand(expr.operator, left, right)
            return float(left) <= float(right)
        elif optype == TokenType.MINUS:
            return float(left) - float(right)
        elif optype == TokenType.PLUS:  # REVIEW: 对于 + 运算符，动态检测出操作数的类型并选择合适的操作
            if isinstance(left, float) and isinstance(right, float):
                return float(left) + float(right)  # 数学运算
            if isinstance(left, str) and isinstance(right, str):
                return str(left) + str(right)  # 连接字符串
            raise LoxRuntimeError(expr.operator, "Operator adds must be two number or two string.")
        elif optype == TokenType.SLASH:
            self.checkNumberOperand(expr.operator, left, right)
            return float(left) / float(right)
        elif optype == TokenType.STAR:
            self.checkNumberOperand(expr.operator, left, right)
            return float(left) * float(right)

        # Unreachable
        return None

    def visitCallExpr(self, expr: Expr.Call):
        callee: object = self.evaluate(expr.callee)  # 可调用可以是任何对象

        arguments = []
        for argument in expr.arguments:
            arguments.append(self.evaluate(argument))

        if not isinstance(callee, LoxCallable):
            raise LoxRuntimeError(expr.paren, "Can only call function and classes.")

        function: LoxCallable = callee  # 调用可调用对象
        if len(arguments) != function.arity():
            raise LoxRuntimeError(expr.paren, f"Expected {function.arity()} arguments but got {len(arguments)} .")

        return function.call(self, arguments)

    def evaluate(self, expr: Expr.Expr):
        """deliver the overriding visitor to expression#accept()"""
        return expr.accept(self)

    def execute(self, stmt: Stmt.Stmt):
        stmt.accept(self)

    def visitBlockStmt(self, stmt: Stmt.Block):
        self.executeBlock(stmt.statements, Environment(self.environment))
        return None

    def executeBlock(self, statements, environment: Environment):
        previous = self.environment
        try:
            self.environment = environment
            for statement in statements:
                self.execute(statement)
        finally:
            self.environment = previous

    def visitExpressionStmt(self, stmt: Stmt.Expression):
        self.evaluate(stmt.expression)
        return None

    def visitFunctionStmt(self, stmt: Stmt.Function):
        function_ = LoxFunction(stmt, self.environment)
        self.environment.define(stmt.name.lexeme, function_)
        return None

    def visitIfStmt(self, stmt: Stmt.If):
        if self.isTruthy(self.evaluate(stmt.condition)):
            self.execute(stmt.thenBranch)
        elif stmt.elseBranch is not None:
            self.execute(self.evaluate(stmt.elseBranch))
        return None

    def visitPrintStmt(self, stmt: Stmt.Print):
        value = self.evaluate(stmt.expression)
        print(self.stringify(value))
        return None

    def visitReturnStmt(self, stmt: Stmt.Return):
        value = None
        if stmt.value is not None:
            value = self.evaluate(stmt.value)
        raise Return(value)

    def visitVarStmt(self, stmt: Stmt.Var):
        value = None
        if stmt.initializer != None:
            value = self.evaluate(stmt.initializer)

        self.environment.define(stmt.name.lexeme, value)
        return None

    def visitWhileStmt(self, stmt: Stmt.While):
        while self.isTruthy(self.evaluate(stmt.condition)):
            self.evaluate(stmt.body)
        return None

    def visitAssignExpr(self, expr: Expr.Assign):
        value = self.evaluate(expr.value)

        self.environment.assign(expr.name, value)
        return value

    def isTruthy(self, obj: object) -> bool:
        """nil and false is truthy"""
        if obj == None: return False  # nil
        if isinstance(obj, bool): return bool(obj)
        return True

    def isEqual(self, a: object, b: object):
        # nil is only equal to nil
        if a is None and b is None: return True
        if a is None: return False

        return a.__eq__(b)

    def stringify(self, obj: object):
        if obj == None: return 'nil'

        if isinstance(obj, float):
            text = str(obj)
            if text.endswith('.0'):
                return text[0:len(text) - 2]  # REVIEW: 不要误用 rstrip，会导致末尾的 0 丢失

        return str(obj)

    def interpreter(self, statements: List[Stmt.Stmt]):
        try:
            for statement in statements:
                self.execute(statement)
        except LoxRuntimeError as error:
            runtimeError(error)



from . import ErrorState
from .Tokens import Token


class LoxRuntimeError(RuntimeError):
    def __init__(self, token: Token, message: str):
        super().__init__(message)
        self.token = token

        self.message = message


def runtimeError(error: LoxRuntimeError):
    print("{}\n[line {}]".format(error.message, error.token.line))
    ErrorState.hadRuntimeError = True



from enum import Enum, auto


class TokenType(Enum):
    LEFT_PAREN, RIGHT_PAREN, LEFT_BRACE, RIGHT_BRACE = [auto() for i in range(4)]
    COMMA, DOT, MINUS, PLUS, SEMICOLON, SLASH, STAR = [auto() for i in range(7)]

    # One or two character tokens.
    BANG, BANG_EQUAL = [auto() for i in range(2)]
    EQUAL, EQUAL_EQUAL = [auto() for i in range(2)]
    GREATER, GREATER_EQUAL = [auto() for i in range(2)]
    LESS, LESS_EQUAL, = [auto() for i in range(2)]

    # Literals.
    IDENTIFIER, STRING, NUMBER, = [auto() for i in range(3)]

    # Keywords.
    AND, CLASS, ELSE, FALSE, FUN, FOR, IF, NIL, OR, = [auto() for i in range(9)]
    PRINT, RETURN, SUPER, THIS, TRUE, VAR, WHILE, = [auto() for i in range(7)]

    EOF = auto()

    # REVIEW: 这里犯了一个非常糟糕的错误，[auto()] 产生不能区分数值的列表，导致标识符完全无法区分


class Token:

    def __init__(self, type: TokenType, lexeme: str, literal: object, line: int):
        self.type, self.lexeme, self.literal, self.line = type, lexeme, literal, line

    def __str__(self):
        return '%s %s %s' % (self.type, self.lexeme, self.literal)

    def __repr__(self):
        args = f'{self.type}, {self.lexeme}, {self.literal}, {self.line}'
        return f'{self.__class__.__name__}({args})'



from .LoxRuntimeError import LoxRuntimeError
from .Tokens import Token


class Environment:
    def __init__(self, enclosing=None):  # None for global scope's environment
        self.enclosing: Environment = enclosing
        self.values = {}

    def define(self, name: str, value: object):
        self.values[name] = value

    def get(self, name: Token):
        """get name in current scope or outer scope"""
        if name.lexeme in self.values.keys():
            return self.values.get(name.lexeme)
        if self.enclosing is not None:
            return self.enclosing.get(name)
        raise LoxRuntimeError(name, f"Undefined variable '{name.lexeme}' .")

    def assign(self, name: Token, value: object):
        if name.lexeme in self.values:
            self.values[name.lexeme] = value
            return
        if self.enclosing is not None:  # 允许修改外部作用域的变量的值
            self.enclosing.values[name.lexeme] = value
            return
        raise LoxRuntimeError(name, f"Undefined variable '{name.lexeme}'.")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="loxpy",
    version="0.1.1",
    author="largomst",
    author_email="yukater@outlook.com",
    description="A small implemetation of lox with python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/largomst/LoxPy",
    packages=["loxpy"],
    install_requires=[],
    entry_points={"console_scripts": ["loxpy = loxpy.Lox:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from typing import List

from .Tokens import TokenType, Token
from .Expr import *


class AstPrinter(ExprVisitor):
    def print(self, expr: Expr) -> str:
        return expr.accept(self)

    def visitBinaryExpr(self, expr: Binary):
        return self.parenthesize(expr.operator.lexeme, expr.left, expr.right)

    def visitGroupingExpr(self, expr: Grouping):
        return self.parenthesize("group", expr.expression)  # REVIEW: group 需要主动命名

    def visitLiteralExpr(self, expr: Literal):
        return str(expr.value)  # REVIEW: 之前误写成 self.parenthesize(str(expr.value)) Literal 是 terminal

    def visitUnaryExpr(self, expr: Unary):
        return self.parenthesize(expr.operator.lexeme, expr.right)

    def parenthesize(self, name, *exprs: Expr):
        return '( {} {} )'.format(name, ' '.join([expr.accept(self) for expr in exprs]))


def main():
    expression = Binary(
        Unary(
            Token(TokenType.MINUS, "-", None, 1),
            Literal(123)),
        Token(TokenType.STAR, "*", None, 1),
        Grouping(
            Literal(45.67)))

    print(AstPrinter().print(expression))


if __name__ == '__main__':
    main()



import sys
from typing import TypeVar

# static check
from . import Interpreter
from . import ErrorState
from .Scanner import Scanner
from . import Parser

ErrorState.hadError = False
ErrorState.hadRuntimeError = False
T = TypeVar("T")

interpreter = Interpreter.Interpreter()  # REVIEW: keep REPL session


def main():
    if len(sys.argv) > 2:
        print("Usage: plox [script]")
    elif len(sys.argv) == 2:
        runFile(sys.argv[1])
    else:
        runPrompt()


def runFile(path: str):
    run(open(path, "r", encoding="utf-8").read())
    if ErrorState.hadError:
        sys.exit(65)
    if ErrorState.hadRuntimeError:
        sys.exit(70)


def runPrompt():
    bang = ">"
    lines = []
    while True:
        lines.append(input(f"{bang} "))
        if lines[-1].endswith("\\"):
            lines[-1] = lines[-1].rstrip("\\")
            bang = ":"
        else:
            run(" ".join(lines))
            lines = []
            bang = ">"


def run(source: str):
    scanner = Scanner(source)
    tokens = scanner.scanTokens()
    # for token in tokens:
    #     print(token)
    parser = Parser.Parser(tokens)
    statements = parser.parse()

    if ErrorState.hadError:
        return

    interpreter.interpreter(statements)


# Map<String, TokenType>




# usr/bin env python3
# coding: utf-8
from typing import List
from .ScannerError import ScanError
from .Tokens import Token, TokenType


class Scanner:
    def __init__(self, source: str):
        self.source = source
        self.tokens = []  # List<Token>

        self.start = 0
        self.current = 0
        self.line = 1

    def scanTokens(self) -> List[Token]:
        while not self.isAtEnd():
            self.start = self.current
            self.scanToken()
        self.tokens.append(Token(TokenType.EOF, "", None, self.line))
        return self.tokens

    def isAtEnd(self) -> bool:
        return self.current >= len(self.source)

    def scanToken(self):
        c = self.advance()
        if c == '(':
            self.addToken(TokenType.LEFT_PAREN)
        elif c == ')':
            self.addToken(TokenType.RIGHT_PAREN)
        elif c == '{':
            self.addToken(TokenType.LEFT_BRACE)
        elif c == '}':
            self.addToken(TokenType.RIGHT_BRACE)
        elif c == ',':
            self.addToken(TokenType.COMMA)
        elif c == '.':
            self.addToken(TokenType.DOT)
        elif c == '-':
            self.addToken(TokenType.MINUS)
        elif c == '+':
            self.addToken(TokenType.PLUS)
        elif c == ';':
            self.addToken(TokenType.SEMICOLON)
        elif c == '*':
            self.addToken(TokenType.STAR)

        elif c == '!':
            self.addToken(TokenType.BANG_EQUAL if self.match('=') else TokenType.BANG)
        elif c == '=':
            self.addToken(TokenType.EQUAL_EQUAL if self.match('=') else TokenType.EQUAL)
        elif c == '<':
            self.addToken(TokenType.LESS_EQUAL if self.match('=') else TokenType.LESS)
        elif c == '>':
            self.addToken(TokenType.GREATER_EQUAL if self.match('=') else TokenType.GREATER)

        # slash 需要特别处理
        elif c == '/':
            if self.match('/'):
                while self.peek() != '\n' and not self.isAtEnd(): self.advance()
            else:
                self.addToken(TokenType.SLASH)

        # 忽略空白字符
        elif c == ' ':
            pass
        elif c == '\r':
            pass
        elif c == '\t':
            pass
        elif c == '\n':
            self.line += 1

        elif c == '"':
            self.string()
        else:
            # 处理数字
            if self.isDigit(c):
                self.number()
            # 处理关键词，方式和处理 <= 相似
            elif self.isAlpha(c):
                self.identifier()
            else:
                self.error(self, "Unexpected character.")

    def advance(self) -> str:
        """**consume**
        return current char and go to next
         consumes the next character in the source file and returns it"""
        self.current += 1
        return self.source[self.current - 1]

    def addToken(self, token_type: TokenType, literal: object = None):
        text = self.source[
               self.start:self.current]  # REVIEW: self.start:self.current 写成了 self.current:self.current 前缀符号丢失
        self.tokens.append(Token(token_type, text, literal, self.line))

    def match(self, expected):
        if self.isAtEnd(): return False  # 位置
        if self.source[self.current] != expected:  # 匹配
            return False
        self.current += 1
        return True

    def peek(self):  # REVIEW: match 和 peek 的功能的差别, match 消耗字符，peek 否
        """**lookahead**
        look but not consume"""
        if self.isAtEnd(): return '\0'  # Q: 为什么要返回 \0
        return self.source[self.current]

    def string(self):
        while self.peek() != '"' and not self.isAtEnd():  # review: 这里的 while 之前错写为 if 导致直接收一个字符
            if self.peek() == '\n': self.line += 1  # 允许字符串跨行
            self.advance()

        # 未正常终结
        if self.isAtEnd():
            self.error(self, "Unterminated string.")
            return

        # 正常闭合 '"'
        self.advance()
        # 这里去掉了两边的 “
        value = self.source[self.start + 1: self.current - 1]  # review: 为什么是 start+1，start 代表着一个 token 的开始
        self.addToken(TokenType.STRING, value)

    def isDigit(self, c: str):
        return '0' <= c <= '9'

    def number(self):
        """match 12345 or 12345.1234
        but not .123 or 1234."""

        while self.isDigit(self.peek()): self.advance()

        # 处理浮点数部分，消耗 “.”
        if self.peek() == '.' and self.isDigit(self.peekNext()):
            self.advance()

            while self.isDigit(self.peek()): self.advance()

        self.addToken(TokenType.NUMBER,
                      float(self.source[self.start: self.current]))

    def peekNext(self) -> str:
        if self.current + 1 > len(self.source): return '\0'
        return self.source[self.current + 1]

    def identifier(self):
        while self.isAlphaNumeric(self.peek()): self.advance()
        # self.addToken(TokenType.IDENTIFIER)

        # review 判断标识符是否为保留字
        text = self.source[self.start: self.current]
        token_type = keywords.get(text)
        if token_type is None: token_type = TokenType.IDENTIFIER
        self.addToken(token_type)

    def isAlpha(self, c: str):
        return 'a' <= c <= 'z' or 'A' <= c <= 'Z' or c == '_'

    def isAlphaNumeric(self, c):
        return self.isAlpha(c) or self.isDigit(c)

    def error(self,token, message):
        err = ScanError(self.line, message)
        err.report()



keywords = {
    "and": TokenType.AND,
    "class": TokenType.CLASS,
    "else": TokenType.ELSE,
    "false": TokenType.FALSE,
    "for": TokenType.FOR,
    "fun": TokenType.FUN,
    "if": TokenType.IF,
    "nil": TokenType.NIL,
    "or": TokenType.OR,
    "print": TokenType.PRINT,
    "return": TokenType.RETURN,
    "super": TokenType.SUPER,
    "this": TokenType.THIS,
    "true": TokenType.TRUE,
    "var": TokenType.VAR,
    "while": TokenType.WHILE,
}
